#! /bin/usr/perl 
#
# pedpeel_2006.pl : returns optimal peeling way for a pedigree
# use "pedpeel_2006.pl --help" or readme.txt for details
#--------------------initial settings------------------------
$name="pedpeel_2006_4.pl";
die "run $name, --help for details\n" if (@ARGV<1);
$version="0.0.4";
$out=" ";
$sep=",";
$g=55; # number of genotypes

$help_message="
***************************************************
SRTNAM: pedpeel.PL 4

HEADER:
$name $version - returns optimal way of pedigree peeling 
for Elston-Stewart algorithm. Requires input file containing
trait data. To process a pedigree file with marker and trait data,
use pedpeel.pl version 0.0.3. To process a pedigree file without
marker and trait columns, pedpeel.pl version 0.0.2 is 
recommended.

SYNOPSIS:
$name [--option] input_pedigree_file

Control input format options:

[--help]
[--sep INPUT_SEPARATOR] - \",\" by default
[--out OUTPUT_SEPARATOR] - \" \" by default
[--g NUMBER_OF_GENOTYPES] - \"55\" by default
[--m NUMBER_OF_GENOTYPES_IN_MEASURED] - \"4\" by default

Control output options:

[--name OUTPUT_FILE_NAME] - \"iway.dat\" by default

DESCRIPTION:
The program prepares pedigree data for calculation of Elston-Stewarts'
likelihood function. It finds an optimal way to peel a pedigree and 
returns text file containing 7 description arrays.

INPUT FILE
Input_pedigree_file is a pedigree file written in standard format:

First string in file is considered as header and will be read to set 
the order of 'ped', 'id', 'fa', 'mo', 'sex', 'prob', and trait data columns.
Trait column name should be given in 'task.inp' file, as its first string.
Only the trait indicated in 'task.inp' will be analysed and rewritten. 
The 'prob' column is not obligatory.

Each following string in file contains information about one pedigree
member. Each individual should have numerical identifier equal to 
the string index number minus 1 (because the first string is a header). 
Parents should be written before their offspring. If these requirements
are not fulfilled, program will stop with error message. 

Pedigree data can be brought to the standard format by recode_ped_2006.pl 
software (http://mga.bionet.nsc.ru/soft/index.html).

'ped' column contains pedigree identifier for each 
	individual. Parts of one pedigree are allowed to be disconnected.
'id' column contains unique numerical identifier 
	for each individual, all individuals should be numbered continuously 
	through input file;
'fa' column contains father's id number for each
	individual, fa=0 if father is unknown;
'mo' column contains mother's id number for each 
	individual, mo=0 if mother is unknown;
'sex' column contains sex code of each individual;
'prob' column contains proband status code of each 
	individual; If the column is absent, proband status of each 
	individual will be assigned	by default as '2'.
Trait data will be rewritten without any changes.

Genotyping status will be assigned from the trait: individuals with 
trait value available will be considered as genotyped. 
Genotyping status will be used for optimisation of the peeling way.
Missing trait data should be coded as '-999'.

Input separator is initially defined as=',' and may be changed by 
[--sep] option. 

PROGRAM BLOCKS
The program finds an optimal way to peel a pedigree having no loops
in structure. 
If pedigree in data file has a loop, program stops with error message.
 STEP 0
Assigning initial settings, opening file 'input_pedigree_file' or 
stop with error message.
 STEP 1
Rewriting data of first pedigree as an array of nuclear 
families.
 STEP 2
Finding an optimal peeling way. 
It is assumed that finish point of optimal peeling way should be 
a nuclear family, in which both parents are founders, i.e. their 
parents are unknown. In this case, the number of possible ways 
corresponds to the number of such nuclear families. The program 
calculates the weight of each way and selects the way with minimal 
weight. The weight will determine running time while calculating
likelihood function. Total peeling weight is a sum of peeling weights
of all nuclear families peeled within pedigree. For each nuclear
family, its peeling weight depends on peeling direction, offspring 
number and number of genotypes possible for family members:
Gf*Gm*(Goff1+Goff2+...+Goffk) when peeling to parent,
GoffP*Gf*Gm*(Goff1+Goff2+...+Goff[k-1]+1) when peeling to offspring,
where 
Gf,Gm,Goff[1..k]-number of genotypes possible for father, mother
and each offspring, respectively;
k - offspring number of the nuclear family;
GoffP - number of genotypes possible for an offspring pointer.
Numbers of possible genotypes are assigned from genotyping status 
induced from marker data. 
Initially, number of genotypes is 55 for individuals with unknown 
genotypes, and 4 for genotyped individuals.
It can be modified by --g and --m options, respectively.
If there are more pedigrees in the input file, the program begins 
processing new pedigree from STEP 1.
 STEP 3
Output peeling sequences in output files.

OUTPUT FILES
The program returns files 'iway.dat' (the name can be changed by
--name option), and '(trait_name)'.

'iway.dat' contains 7 description arrays written as 7 strings.
First three strings are INS, IPS and INP arrays containing accumulated 
sizes of pedigrees, numbers of nuclear families in pedigrees and 
sizes of these nuclear families, respectively.

The arrays consist from following elements:
1) [the number of structures];
2) [0];
3) [number of elements in the first structure];
4) [number of elements in the 1st and 2nd structures];
...
i+2) [number of elements accumulated in first i structures];
...
n+2) [number of elements accumulated in all n structures] - the last
array element.

'structure' is a pedigree in both INS and IPS, and a nuclear family in INP.
'elements' are individuals in both INS and INP, and nuclear families in IPS.
Nuclear families in INP array are written in the order they will be peeled. 
First number in 1st string is repeated first element of INS array.
First number in 3rd string is the number of INP array elements.

4th INF array is that of ids of nuclear family members and has 
the following structure: 
[fathers' id of 1st family] [mothers' id of 1st family] 
[id of first offspring of 1st family] ... 
[id of last offspring of 1st family] [fathers' id of 2nd family] ... 
[id of last offspring of last family]. 
'id' is calculated as 'initial id' minus number of ids accumulated  
in previous pedigrees.
First number in 4th string is that of INF array elements. 
Nuclear families are written in peeling order. If nuclear family is 
to be peeled to an offspring, id of the offspring will be written 
just after mothers' id.

5th IWAY array is that of pointers' ids written in peeling order.
'id' is calculated as 'initial id' minus number of ids accumulated  
in previous pedigrees.

6th ISEX array contains sex codes for each individual ordered as they
are written in input file. First number in 6th string is that of 
individuals in the sample.

7th IPROB array contains 'prob' codes for each individual ordered as they
are written in input file.

'(trait_name)' contains one string where trait phenotypes are 
written for each individual (ordered as in input file).

All warning messages and other information will be written in file 
'inform.txt'.

RESOURCES:

DIAGNOSTICS:

SEE ALSO:

EXAMPLE:
pedpeel_2006_4.pl --help
pedpeel_2006_4.pl pedpeel4_example.csv
pedpeel_2006_4.pl pedpeel4_example.csv --g 4 --out \"\-\" --name output.txt 

COPYRIGHT:
pedpeel.pl is Copyright (c) 2004 Erasmus MC Rotterdam and ICG Novosibirsk
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 2 of the License, or (at your opinion) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

BUGS:

AUTHOR:
Nadezhda Belonogova (adg101\@mail.ru)

VERSION:
$version 2006

CREATED:
19.11.2008

LAST MODIFIED:
19.11.2008

LANGUAGE:
Perl v5.8.4

OS: 
XXX Linux vXX.XX

***************************************************\n\n";

#================ STEP 0: initial settings ==============

if ($ARGV[0] eq "--help") {
	open RDM, ">readme.txt";
	print RDM "${help_message}";
	print "${help_message}";
	print "'readme.txt' file containing the help_message has been created\n";
	close RDM;
	die ""; 
} 
if (@ARGV==1) {	
	$pedfile=$ARGV[0];
	} 
else {	
	$narg = @ARGV;
	$pedfile=$ARGV[0];
	for ($i=1;$i<$narg;$i+=2) 
	{
		$sep=$ARGV[$i+1] if ($ARGV[$i] =~ /--sep/);
		$out=$ARGV[$i+1] if ($ARGV[$i] =~ /--out/);
		if ($ARGV[$i] =~ /--g/){
			$g=$ARGV[$i+1];
			if ($g<4){
				$er="ERROR: Number of genotypes cannot be less than 4!\n";
				print WRN $er;
				print $er;
				$g=4}}
		if ($ARGV[$i] =~ /--m/){
			$gm=$ARGV[$i+1];
			if ($gm<4){
				$er="ERROR: Number of genotypes cannot be less than 4!\n";
				print WRN $er;
				print $er;
				$gm=4}}
		$filename=$ARGV[$i+1] if ($ARGV[$i] =~ /--name/);
		}}

$gm=4 if (defined$gm eq "");
open PED, $pedfile or die "FATAL ERROR: Cannot open file $pedfile!\n";
$filename="iway.dat" if (defined$filename eq "");

open WRN, ">inform.txt";
open MAS, ">${filename}";
#-------------------TASK---------------------
open TASK, "<task.inp" or die "Cannot open 'task.inp' file!";
$task=<TASK>;
@task = split(/\s/, $task);
$phen=$task[0];
#--------------------------------------------
$tim=localtime;
$er="Run $name v. $version
\t${tim}\nInput parameters:
\tField input separator: --sep='$sep'
\tField output separator: --out='$out'
\tNumber of genotypes: --g='$g'
\tNumber of genotypes in measured: --m='$gm'
\tInput pedigree file: '$pedfile'
\tOutput pedigree file: --name='$filename'
\tTrait to be analyzed: '$phen'\n";

print WRN $er;
print $er;

$part=1;
$jinp[0]=0;
$ips[0]=0;
$ins[0]=0;
$inpcount=0;
$infcount=0;
$ips=0;
$line=<PED>; 

@header = split(/$sep/, $line);#header processed
$enmar=0;
for ($i=0;$i<@header;$i++) {
	if ($header[$i] =~ /^id$/){	$enid++; 	$id=$i}
	if ($header[$i] =~ /^ped$/){	$enped++;	$pd=$i}
	if ($header[$i] =~ /^fa$/){	$enfa++;	$fa=$i}
	if ($header[$i] =~ /^mo$/){	$enmo++;	$mo=$i}
	if ($header[$i] =~ /^sex$/){	$ensex++;	$sx=$i}
	if ($header[$i] =~ /^prob$/){	$enprob++;	$pb=$i}
	if ($header[$i] =~ /^${phen}$/){	$enph++;	$ph=$i}}

$er=0;
if ($enid!=1){$er[$er]="'id'"; $er++}
if ($enped!=1){$er[$er]="'ped'"; $er++}
if ($enfa!=1){$er[$er]="'fa'"; $er++}
if ($enmo!=1){$er[$er]="'mo'"; $er++}
if ($ensex!=1){$er[$er]="'sex'"; $er++}
if ($enph!=1){$er[$er]="'${phen}'"; $er++}
if (@er>0){
	$er=join ", ", @er;
	print WRN "FATAL ERROR: Wrong data format: cannot find $er in header!\n";
	die "FATAL ERROR: Wrong data format: cannot find $er in header!\n"}
undef@er;

if ($enprob!=1){
	$er = "WARNING: Cannot find 'prob' in header!\n	Proband status will be assigned as '2' by default!\n";
	print $er;
	print WRN $er;
	undef$pb}
	
$lnum=0;
$ped=0;
$ped[$ped]=0;
$line=<PED>; # first string
@temp = split (/\n/, $line);
@temp = split (/$sep/, $temp[0]);

READ: 
while ($line ne "") {
	$i=0; # NF number
	$ped++;
	$pedid=$lnum;
	$fid=0;
	undef@fid;
	if ($temp[$pd] ne ""){
		foreach (@ped) {#check for pedigree breaks
			if ($_==$temp[$pd]){
				$lnum++;
				$er="FATAL ERROR: Wrong data format: Pedigree $temp[$pd] is broken into two blocks at line $lnum!\n";
				print WRN $er;
				die "$er";
				}}
		$ped[$ped]=$temp[$pd];}# array of pedigree numbers
	else {
		$lnum++;
		$er="FATAL ERROR: Pedigree number is absent at line $lnum!\n";	
		print WRN $er;
		die "$er";}  
	
#============ STEP 1: rewriting pedigree data ==================
	$tim=localtime;
	$stim=substr($tim,11,8);
	print WRN "\nPedigree $ped: $stim\n";
	print "Pedigree $ped:\nSTEP 1: rewriting pedigree data...\n$stim\n";
	
	while ($ped[$ped]==$temp[$pd]) {
		$lnum++;
		
		if ($temp[$id]!=$lnum){
			$er="FATAL ERROR: Wrong data format: id $temp[$id] is not equal to string number at line $lnum!\n" ;
			print WRN $er;
			die "$er"}
			
		if ($temp[$id]<=$temp[$fa] or $temp[$id]<=$temp[$mo]){
			$er="FATAL ERROR: Wrong data format: offspring $temp[$id] is written before parent at line $lnum!\n" ;
			print WRN $er;
			die "$er"}
			
		if (($temp[$fa]<=$pedid and $temp[$fa]!=0) or ($temp[$mo]<=$pedid and $temp[$mo]!=0)){
			$er="FATAL ERROR: Parent at line $lnum is absent in pedigree $ped!\n" ;
			print WRN $er;
			die "$er"}
			
		if ($temp[$sx]==1 or $temp[$sx]==2){$sex[$lnum-1]=$temp[$sx];}
		else {
			$er="FATAL ERROR: Wrong sex code at line $lnum!\n";
			print WRN $er;
			die "$er"}

		if (defined$pb ne ""){
			if ($temp[$pb] ne ""){$prob[$lnum-1]=$temp[$pb]}
			else {$prob[$lnum-1]=2}}
		else {$prob[$lnum-1]=2}

		$ph[$lnum-1]=$temp[$ph];

		if ($temp[$ph]ne"-999"){
			$izm[$temp[$id]]=$gm}
		else {
			$izm[$temp[$id]]=$g}

		if (($temp[$fa] == 0 and $temp[$fa] ne "0") or ($temp[$mo] ==0 and $temp[$mo] ne "0") or ($temp[$id] ==0 and $temp[$id] ne "0")){
			$er="FATAL ERROR: Wrong or absent data at line $lnum!\n";
			print WRN $er; 
			die "$er"}

$np{$temp[$id]}=0;

if ($temp[$fa] > 0 or $temp[$mo] > 0) { # if at least one parent is known
	$isearch=$i; # the number of NFs added already
	$ch=0; #add child switch
	$secfath=0; #second fathers' marriage switch
	$secmoth=0; #second mothers' marriage switch
		for ($j=1;$j<=$isearch;$j++){
#-----------------sib search---------------------------
		if ($ch==0){ 
			for ($n=1;$n<=$isearch;$n++){
			if ($temp[$fa] eq $np{"fath$n"} && $temp[$mo] eq $np{"moth$n"}) { # if ids of both parents are the same as in line j
			$off[$n]++;
			$ch=$off[$n];
			$np{"off${ch}_$n"}=$temp[$id];
			goto M1;
		}}
		$i++;
		$np{"fath$i"}=$temp[$fa];
		$np{"moth$i"}=$temp[$mo];
		$np{"off1_$i"}=$temp[$id];
		$off[$i]=1;}
		$ch=1; # no sibs found among added NFs
#-----------------second marriage check---------------------------			
				if ($temp[$fa] == $np{"fath$j"} and $temp[$mo] ne $np{"moth$j"} and $secfath == 0){ #if fathers' id is the same - the second marriage
				$secfath=1;
				$mar{$temp[$fa]}++;
				$np{$temp[$fa]}++; #adding the new connection for father
				}
				if ($temp[$mo] == $np{"moth$j"} and $temp[$fa] ne $np{"fath$j"} and $secmoth == 0){ #if mothers' id is the same - the second marriage
				$secmoth=1;
				$mar{$temp[$mo]}++;
				$np{$temp[$mo]}++;#adding the new connection for mother
				}
#------------------grandchildren check------------------------------
	for ($k=1; $k<=$off[$j]; $k++){#search parents among the offspring
		if ($np{"off${k}_$j"}==$temp[$fa] and $grch[$temp[$fa]]!=1){#father case
		$grch[$temp[$fa]]=1;
		$np{$temp[$fa]}++;
		}
		if ($np{"off${k}_$j"}==$temp[$mo] and $grch[$temp[$mo]]!=1){#mother case
		$grch[$temp[$mo]]=1;
		$np{$temp[$mo]}++;}
	}}
		
#-------------------other cases---------------------------
	if ($i==0){ #first NP
		$i++;
		$np{"fath$i"}=$temp[$fa];
		$np{"moth$i"}=$temp[$mo];
		$np{"off1_$i"}=$temp[$id];
		$off[$i]=1;}
	}
else {
	$fid++;
	$fid[$fid]=$temp[$id]}
M1:
$line=<PED>;
@temp = split (/\n/, $line);
@temp = split (/$sep/, $temp[0]);
}
$ins[$ped]=$lnum;
$ips=$ips+$i;
$ips[$ped]=$ips;
print "Pedigree rewriting finished\n";
if ($i==0 and $fid!=0){
	print WRN "WARNING: Pedigree $ped contains disconnected individuals only!\n";
	next READ} 

#================== STEP 2: pedpeel =========================
$tim=localtime;
$stim=substr($tim,11,8);
print "STEP 2: search for optimal peeling way...\n$stim\n";
$num=1; # define$num: NF number peeled
@rest=(0..$i);
$jfc=0;
@jf=0;
for ($j=1; $j<=$i;$j++){ # define$j: cycle name for founders' NFs
	$fathid=$np{"fath$j"};
	$mothid=$np{"moth$j"};
	$mothcon=$np{"$mothid"}-$mar{"$mothid"};
	$fathcon=$np{"$fathid"}-$mar{"$fathid"};
	if ($mothcon==0 and $fathcon==0){ # if it is founders' NF
		$jf[$jfc]=$j;
		$jfc++;}
	if ($mothcon==0 or $fathcon==0){
		for ($f=1; $f<=$fid; $f++) {
			delete($fid[$f]) if ($fid[$f]==$fathid or $fid[$f]==$mothid)}}}
if (@fid!=0){
	print WRN "WARNING: Pedigree $ped contains both disconnected individuals and nuclear families!\n"}
JF:
$numpart++;
$minsum=0;
@jfold=@jf;
@jfpeel=0;
$jfc=1;
$jfpeel[0]=$jf[0];

foreach (@jfpeel){
	$J=$peel[1]=$_; # define@peel: array of ids of peeled NF
	$offnum=$off[$J];
	$up[$offnum]++; # define@up: array of numbers of different child numbers in peeled up NF
	$mothid=$np{"moth$J"};
	$mG=$izm[$mothid];
	$fathid=$np{"fath$J"};
	$fG=$izm[$fathid];
	$item=0;
	for ($k=1; $k<=$off[$J]; $k++){ 
		$offid=$np{"off${k}_$J"};
		$ofG=$izm[$offid];
		$item=$item+$ofG;}
		$item=$item*$mG*$fG;
	$sum=$item;
	for ($n=1; $n<=$num; $n++){ # \# families peeled, define$n: cycle name for number of NF peeled in this cycle
		$J=$peel[$n];
		$rest[$J]=0;
		$mothid=$np{"moth$J"};
		$fathid=$np{"fath$J"};
		$mothcon=$np{"$mothid"};
		$fathcon=$np{"$fathid"};
		$pnum++;
		if ($fathcon>0 and $fathid!=$pointer[$n]) {
			$poinid[$pnum]=$fathid;
			$poincon[$pnum]=$fathcon;}
		$pnum++;
		if ($mothcon>0 and $mothid!=$pointer[$n]) {
			$poinid[$pnum]=$mothid;
			$poincon[$pnum]=$mothcon;}
		for ($k=1; $k<=$off[$J]; $k++){ 
			$offid=$np{"off${k}_$J"};
			$offcon=$np{"$offid"};
			if ($offcon>0 and $offid!=$pointer[$n]){ #if the child is connected then write his id and con
				$pnum++;
				$poinid[$pnum]=$offid;
				$poincon[$pnum]=$offcon}
		}
#--------------search new NF to peel-------------------					
SEARCH:	
		for ($Jc=1; $Jc<=@rest-1; $Jc++){ # search among rest NF
		$J=$rest[$Jc];
		next SEARCH if ($J==0); # excluding family already processed
		for ($p=1;$p<=$pnum;$p++){ # for each new pointer
			if ($n>1 and $p<3){ # search among offspring, for parents only
				for ($k=1; $k<=$off[$J]; $k++){ 
					if ($np{"off${k}_$J"}==$poinid[$p]){ #family to peel to an offspring
						for ($jch=$num; $jch>=1; $jch--){ # excluding families which are already found
							if ($J==$peel[$jch]){
								$er="FATAL ERROR: Loop found while processing pedigree $ped!\n";
								print WRN $er;
								die "$er"}}
						$num++;
						$peel[$num]=$J;
						$pointer[$num]=$poinid[$p];
						$offnum=$off[$J];
						$down[$offnum]++;
#$sum counting, down case:
						$mothid=$np{"moth$J"};
						$mG=$izm[$mothid];
						$fathid=$np{"fath$J"};
						$fG=$izm[$fathid];
						$item=0;
						if ($offnum>1){
							for ($kk=1; $kk<=$off[$J]; $kk++){ 
								$offid=$np{"off${kk}_$J"};
									if ($k==$kk){$pG=$izm[$offid]}
									else {
										$ofG=$izm[$offid];
										$item=$item+$ofG}}}
						else {$pG=$izm[$offid];}
						$item++;
						$item=$item*$mG*$fG*$pG;
						$sum=$sum+$item;
						if ($_==$jf[0]){
						$mothcon=$np{"$mothid"}-$mar{"$mothid"};
						$fathcon=$np{"$fathid"}-$mar{"$fathid"};
						if ($mothcon==0 and $fathcon==0){ #rewriting array of founders' NF
							$jfpeel[$jfc]=$J;
							$jfc++}}
						$fI++;
						$forINF[$fI]=$J;
						$kINF[$fI]=$k;
						goto M5 if ($minsum>0 and $sum>$minsum);
						goto M3}}}
M3:		if ($np{"fath$J"}==$poinid[$p] or $np{"moth$J"}==$poinid[$p]){#search among parents (for all families)
				for ($jch=$num; $jch>=1; $jch--){ # excluding families which are already found 
					if ($J==$peel[$jch]){
						$er="FATAL ERROR: Loop found while processing pedigree $ped!\n";
						print WRN $er;
						die "$er"}}
				$num++;
				$peel[$num]=$J;
				$pointer[$num]=$poinid[$p];
				$offnum=$off[$J];
				$up[$offnum]++; # define@up: array of numbers of different child numbers in peeled up NF
#$sum counting, up case:
				$mothid=$np{"moth$J"};
				$mG=$izm[$mothid];
				$fathid=$np{"fath$J"};
				$fG=$izm[$fathid];
				$item=0;
				for ($k=1; $k<=$off[$J]; $k++){ 
					$offid=$np{"off${k}_$J"};
					$ofG=$izm[$offid];
					$item=$item+$ofG;}
				$item=$item*$mG*$fG;
				$sum=$sum+$item;
				if ($_==$jf[0]){
					$mothcon=$np{"$mothid"}-$mar{"$mothid"};
					$fathcon=$np{"$fathid"}-$mar{"$fathid"};
					if ($mothcon==0 and $fathcon==0){ #rewriting array of founders' NF
						$jfpeel[$jfc]=$J;
						$jfc++}}
				goto M5 if ($minsum>0 and $sum>$minsum)}}}
		@poinid=0;
		@poincon=0;
		$pnum=0;}
	$minsum=$sum;
	$jfinish=$_;
	@mforINF=@forINF;
	@mkINF=@kINF;
	@mpeel=@peel;
	@mpoin=@pointer;
	@mup=@up;
	@mdown=@down;
#-----------------restoring all values--------------------
	M5: 
	@poinid=0;
	@poincon=0;
	$pnum=0;
	@peel=0;
	@pointer=0;
	@down=0;
	@up=0;
	$num=1;
	@rest=(0..$i);
	$sum=0;
	@forINF=0;
	@kINF=0;
	$fI=0;
	goto M20 if ($minsum==64);
	M4:}

#--------------counting real order-----------------------
M20:
$rew=0;
@rfam=0;
@rpoin=0;
for ($r=@mpeel-1; $r>=2; $r--){ #rewriting of peeling order
	$rew++; # new position
	$rfam[$rew]=$mpeel[$r]; # define@rfam: array of NF ids already peeled
	$rpoin[$rew]=$mpoin[$r];
	}

#-----------------------output arrays--------------------------

#--------------INP array---------------
for ($r=1; $r<=(@rfam-1); $r++){
	$J=$rfam[$r];
	$inp=$inp+$off[$J]+2;
	$inp[$r-1]=$inp;
#------------INF array-------------
	$mothid=$np{"moth$J"}-$ins[$ped-1];
	$fathid=$np{"fath$J"}-$ins[$ped-1];
	$inf="$inf$out$fathid$out$mothid";
	for ($fI=@mforINF-1; $fI>=1; $fI--){
		$Jd=$mforINF[$fI];
		if ($Jd==$J){
			$ofpoin=$rpoin[$r]-$ins[$ped-1];
			$inf="$inf$out$ofpoin";
K:		for ($k=1;$k<=$off[$J];$k++){
				next K if ($k==$mkINF[$fI]);
				$offid=$np{"off${k}_$J"}-$ins[$ped-1];
				$inf="$inf$out$offid"}
			goto M10}}
	
	for ($k=1;$k<=$off[$J];$k++){
		$offid=$np{"off${k}_$J"}-$ins[$ped-1];
		$inf="$inf$out$offid"}
#--------------IWAY array---------------
M10: 
$iway[$r-1]=$rpoin[$r]-$ins[$ped-1];
}
$iway[$r-1]=0;
$inp=$inp+$off[$jfinish]+2;
$inp[$r-1]=$inp;
$jinp = join "$out", @inp;
$jinp[$part]=$jinp;
@inp=0;
$mothid=$np{"moth$jfinish"}-$ins[$ped-1];
$fathid=$np{"fath$jfinish"}-$ins[$ped-1];
$inf="$inf$out$fathid$out$mothid";
for ($k=1;$k<=$off[$jfinish];$k++){
	$offid=$np{"off${k}_$jfinish"}-$ins[$ped-1];
	$inf="$inf$out$offid";}
$jinf[$part]=$inf;
undef$inf;
$jiw = join "$out", @iway;
$jiw[$part-1]=$jiw; 
@iway=0;
$part++;

#-----------------rewriting founders' NF array---------------
if (@jfold!=@jfpeel){
@jf=0;
$jfc=0;
REW:
for ($r=0; $r<@jfold; $r++){
	foreach (@jfpeel){
		next REW if ($jfold[$r]==$_)}
		$jf[$jfc]=$jfold[$r];
		$jfc++}
	if (@jf>=1){
		$mpeel=@mpeel-1;
		$pl="families" if ($mpeel>1);
		$pl="family" if ($mpeel==1);
		$er="Part $numpart:\n$mpeel nuclear $pl peeled with minimal weight = $minsum units\n";
		print $er;
		print WRN $er;
		goto JF}}
if ($numpart==1){
	$pl="families" if ($i>1);
	$pl="family" if ($i==1);	
	$er="$i nuclear $pl peeled with minimal weight = $minsum units\n";
	print $er;
	print WRN $er;}
else {
	$mpeel=@mpeel-1;
	$pl="families" if ($mpeel>1);
	$pl="family" if ($mpeel==1);
	$er="Part $numpart:\n$mpeel nuclear $pl peeled with minimal weight = $minsum units\n";
	print $er;
	print WRN $er;}
$numpart=0;
$mins=$mins+$minsum;
} #next pedigree

close PED;
close OUT;

#=========== STEP 3: OUTPUT ==============

#--------------INS----------------
print MAS "$ped$out$ped"; # ids accumulated in pedigrees
foreach (@ins){
print MAS "$out$_";}
print MAS "\n";
#--------------IPS----------------
print MAS "$ped"; # NFs accumulated in pedigrees
foreach (@ips){
print MAS "$out$_";}
print MAS "\n";
#--------------INP---------------
$first=$ips+2;
print MAS "$first$out$ips$out"; # ids accumulated in NFs
foreach (@jinp){
print MAS "$_$out";}
print MAS "\n";
#--------------INF---------------
print MAS "$inp"; # peeling order
foreach (@jinf){
print MAS "$_";}
print MAS "\n";
#--------------IWAY--------------
foreach (@jiw){ # pointers' order
print MAS "$_$out";}
print MAS "\n";
#--------------ISEX--------------
print MAS "$lnum";
foreach (@sex){
print MAS "$out$_";}
print MAS "\n";
#-------------IPROB--------------
foreach (@prob){
print MAS "$_$out";}
print MAS "\n";
close MAS;
#-------------PHEN---------------
open PH, ">${phen}";
foreach (@ph){
print PH "$_$out";}
close PH;

$tim=localtime;
$er="Processing finished successfully!\nTotal weight is $mins units\n$tim\n";
print $er;
print WRN "\n$er";
close WRN; 